local addonName, addonTable = ...; -- Pulls back the Addon-Local Variables and store them locally.
local addon = _G[addonName];
local L = LibStub("AceLocale-3.0"):GetLocale(addonName);

--- Localize Vars
local coreGeneral = addon.Core.General;
local Objects = addon.Core.Objects;
local Settings = addon.Core.Settings;
local val = coreGeneral.ToNumber;

-- Objects
local Pet = addon.Units.Pet;
local Player = addon.Units.Player;
local Target = addon.Units.Target;
local Racial, Spell, Talent, Aura, Azerite, Item, Consumable;

-- Rotation Variables
local nameAPL = "lunaeclipse_mage_frost";

-- AOE Rotation
local function AOE(rotationName)
	-- Inherits Rotation Class so get the base class.
	local self = addonTable.rotationsClass(nameAPL, rotationName);

	-- Table to hold requirements to use spells and items listed in the rotation.
	self.Requirements = {
		-- # Simplified Flurry conditions from the ST action list. Since the mage is generating far less Brain Freeze charges, the exact condition here isn't all that important.
		-- actions.aoe+=/flurry,if=prev_gcd.1.ebonbolt|buff.brain_freeze.react&(prev_gcd.1.frostbolt&(buff.icicles.stack<4|!talent.glacial_spike.enabled)|prev_gcd.1.glacial_spike)
		Flurry = function()
			return Player.PrevGCD(1, Talent.Ebonbolt)
			    or Player.Buff(Aura.BrainFreeze).React()
			   and (Player.PrevGCD(1, Spell.Frostbolt) and (Player.Buff(Aura.Icicles).Stack() < 4 or not Talent.GlacialSpike.Enabled()) or Player.PrevGCD(1, Talent.GlacialSpike));
		end,

		-- SimCraft doesn't specify this, but only cast if not moving.
		-- We need to put the condition here because movement check is not working yet.
		-- actions.aoe+=/frostbolt
		Frostbolt = function()
			return not Player.IsMoving();
		end,

		-- actions.aoe+=/ice_lance,if=buff.fingers_of_frost.react
		IceLance = function()
			return Player.Buff(Aura.FingersOfFrost).React();
		end,
	};

	Objects.FinalizeRequirements(self.Requirements);

	function self.Rotation(action)
		-- # With Freezing Rain, it's better to prioritize using Frozen Orb when both FO and Blizzard are off cooldown. Without Freezing Rain, the converse is true although the difference is miniscule until very high target counts.
		-- actions.aoe=frozen_orb
		action.EvaluateAction(Spell.FrozenOrb, true);
		-- actions.aoe+=/blizzard
		action.EvaluateAction(Spell.Blizzard, true);
		-- actions.aoe+=/comet_storm
		action.EvaluateAction(Talent.CometStorm, true);
		-- actions.aoe+=/ice_nova
		action.EvaluateAction(Talent.IceNova, true);
		action.EvaluateAction(Spell.Flurry, self.Requirements.Flurry);
		action.EvaluateAction(Spell.IceLance, self.Requirements.IceLance);
		-- # The mage will generally be generating a lot of FoF charges when using the AoE action list. Trying to delay Ray of Frost until there are no FoF charges and no active Frozen Orbs would lead to it not being used at all.
		-- actions.aoe+=/ray_of_frost
		action.EvaluateAction(Talent.RayOfFrost, true);
		-- actions.aoe+=/ebonbolt
		action.EvaluateAction(Talent.Ebonbolt, true);
		-- actions.aoe+=/glacial_spike
		action.EvaluateAction(Talent.GlacialSpike, true);
		-- # Using Cone of Cold is mostly DPS neutral with the AoE target thresholds. It only becomes decent gain with roughly 7 or more targets.
		-- actions.aoe+=/cone_of_cold
		action.EvaluateAction(Spell.ConeOfCold, true);
		action.EvaluateAction(Spell.Frostbolt, self.Requirements.Frostbolt);
		-- actions.aoe+=/ice_lance
		action.EvaluateAction(Spell.IceLance, true);
	end

	-- # The target threshold isn't exact. Between 3-5 targets, the differences between the ST and AoE action lists are rather small. However, Freezing Rain prefers using AoE action list sooner as it benefits greatly from the high priority Blizzard action.
	-- actions+=/call_action_list,name=aoe,if=active_enemies>3&talent.freezing_rain.enabled|active_enemies>4
	function self.Use(numEnemies)
		return (numEnemies > 3 and Talent.FreezingRain.Enabled())
		    or numEnemies > 4;
	end

	return self;
end

-- Create a variable so we can call the rotations functions.
local rotationAOE = AOE("AOE");

-- RuneOfPower Rotation
local function RuneOfPower(rotationName)
	-- Inherits Rotation Class so get the base class.
	local self = addonTable.rotationsClass(nameAPL, rotationName);

	-- Table to hold requirements to use spells and items listed in the rotation.
	self.Requirements = {
		RuneOfPower = {
			-- # With Glacial Spike, Rune of Power should be used right before the Glacial Spike combo (i.e. with 5 Icicles and a Brain Freeze). When Ebonbolt is off cooldown, Rune of Power can also be used just with 5 Icicles.
			-- actions.talent_rop=rune_of_power,if=talent.glacial_spike.enabled&buff.icicles.stack=5&(buff.brain_freeze.react|talent.ebonbolt.enabled&cooldown.ebonbolt.remains<cast_time)
			GlacialSpike = function()
				return Talent.GlacialSpike.Enabled()
				   and Player.Buff(Aura.Icicles).Stack() == 5
				   and (Player.Buff(Aura.BrainFreeze).React() or Talent.Ebonbolt.Enabled() and Talent.Ebonbolt.Cooldown.Remains() < Talent.RuneOfPower.CastTime());
			end,

			-- # Without Glacial Spike, Rune of Power should be used before any bigger cooldown (Ebonbolt, Comet Storm, Ray of Frost) or when Rune of Power is about to reach 2 charges.
			-- actions.talent_rop+=/rune_of_power,if=!talent.glacial_spike.enabled&(talent.ebonbolt.enabled&cooldown.ebonbolt.remains<cast_time|talent.comet_storm.enabled&cooldown.comet_storm.remains<cast_time|talent.ray_of_frost.enabled&cooldown.ray_of_frost.remains<cast_time|charges_fractional>1.9)
			Use = function()
				return not Talent.GlacialSpike.Enabled()
				   and (Talent.Ebonbolt.Enabled() and Talent.Ebonbolt.Cooldown.Remains() < Talent.RuneOfPower.CastTime() or Talent.CometStorm.Enabled() and Talent.CometStorm.Cooldown.Remains() < Talent.RuneOfPower.CastTime() or Talent.RayOfFrost.Enabled() and Talent.RayOfFrost.Cooldown.Remains() < Talent.RuneOfPower.CastTime() or Talent.RuneOfPower.Charges.Fractional() > 1.9);
			end,
		},
	};

	Objects.FinalizeRequirements(self.Requirements);

	function self.Rotation(action)
		action.EvaluateAction(Talent.RuneOfPower, self.Requirements.RuneOfPower.GlacialSpike);
		action.EvaluateAction(Talent.RuneOfPower, self.Requirements.RuneOfPower.Use);
	end

	-- # On single target fights, the cooldown of Rune of Power is lower than the cooldown of Frozen Orb, this gives extra Rune of Power charges that should be used with active talents, if possible.
	-- actions.cooldowns+=/call_action_list,name=talent_rop,if=talent.rune_of_power.enabled&active_enemies=1&cooldown.rune_of_power.full_recharge_time<cooldown.frozen_orb.remains
	function self.Use(numEnemies)
		return Talent.RuneOfPower.Enabled()
		   and numEnemies == 1
		   and Talent.RuneOfPower.Charges.FullRechargeTime() < Spell.FrozenOrb.Cooldown.Remains();
	end

	return self;
end

-- Create a variable so we can call the rotations functions.
local rotationRuneOfPower = RuneOfPower("RuneOfPower");

-- Cooldowns Rotation
local function Cooldowns(rotationName)
	-- Inherits Rotation Class so get the base class.
	local self = addonTable.rotationsClass(nameAPL, rotationName);

	-- Table to hold requirements to use spells and items listed in the rotation.
	self.Requirements = {
		-- actions.cooldowns+=/potion,if=prev_gcd.1.icy_veins|target.time_to_die<70
		BattlePotionOfIntellect = function()
			return Player.PrevGCD(1, Spell.IcyVeins)
			    or Target.TimeToDie() < 70;
		end,

		-- # Rune of Power is always used with Frozen Orb. Any leftover charges at the end of the fight should be used, ideally if the boss doesn't die in the middle of the Rune buff.
		-- actions.cooldowns+=/rune_of_power,if=prev_gcd.1.frozen_orb|time_to_die>10+cast_time&time_to_die<20
		RuneOfPower = function()
			return Player.PrevGCD(1, Spell.FrozenOrb)
			    or Target.TimeToDie() > 10 + Talent.RuneOfPower.CastTime()
			   and Target.TimeToDie() < 20;
		end,
	};

	Objects.FinalizeRequirements(self.Requirements);

	function self.Rotation(action)
		-- actions.cooldowns+=/icy_veins
		action.EvaluateAction(Spell.IcyVeins, true);
		-- actions.cooldowns+=/mirror_image
		action.EvaluateAction(Talent.MirrorImage, true);
		action.EvaluateAction(Talent.RuneOfPower, self.Requirements.RuneOfPower);

		action.CallActionList(rotationRuneOfPower);

		action.EvaluateAction(Consumable.BattlePotionOfIntellect, self.Requirements.BattlePotionOfIntellect);
		-- actions.cooldowns+=/blood_fury
		action.EvaluateAction(Racial.BloodFury, true);
		-- actions.cooldowns+=/berserking
		action.EvaluateAction(Racial.Berserking, true);
		-- actions.cooldowns+=/lights_judgment
		action.EvaluateAction(Racial.LightsJudgment, true);
		-- actions.cooldowns+=/fireblood
		action.EvaluateAction(Racial.Fireblood, true);
		-- actions.cooldowns+=/ancestral_call
		action.EvaluateAction(Racial.AncestralCall, true);
	end

	return self;
end

-- Create a variable so we can call the rotations functions.
local rotationCooldowns = Cooldowns("Cooldowns");

-- SingleTarget Rotation
local function SingleTarget(rotationName)
	-- Inherits Rotation Class so get the base class.
	local self = addonTable.rotationsClass(nameAPL, rotationName);

	-- Table to hold requirements to use spells and items listed in the rotation.
	self.Requirements = {
		Blizzard = {
			-- # With Freezing Rain and at least 2 targets, Blizzard needs to be used with higher priority to make sure you can fit both instant Blizzards into a single Freezing Rain. Starting with three targets, Blizzard leaves the low priority filler role and is used on cooldown (and just making sure not to waste Brain Freeze charges) with or without Freezing Rain.
			-- This doesn't make sense, I think buff.fingers_of_frost.react<2 should be Remains < 2
			-- actions.single+=/blizzard,if=active_enemies>2|active_enemies>1&cast_time=0&buff.fingers_of_frost.react<2
			FingersOfFrost = function(numEnemies)
				return numEnemies > 2
				    or numEnemies > 1
				   and Spell.Blizzard.CastTime() == 0
				   and Player.Buff(Aura.FingersOfFrost).Remains() < 2;
			end,

			-- # Blizzard is used as low priority filler against 2 targets. When using Freezing Rain, it's a medium gain to use the instant Blizzard even against a single target, especially with low mastery.
			-- actions.single+=/blizzard,if=cast_time=0|active_enemies>1
			Use = function(numEnemies)
				return Spell.Blizzard.CastTime() == 0
				    or numEnemies > 1;
			end,
		},

		-- # Without GS, Ebonbolt is used on cooldown. With GS, Ebonbolt is only used to fill in the blank spots when fishing for a Brain Freeze proc, i.e. the mage reaches 5 Icicles but still doesn't have a Brain Freeze proc. This is DPS neutral with low mastery and slight gain otherwise.
		-- actions.single+=/ebonbolt,if=!talent.glacial_spike.enabled|buff.icicles.stack=5&!buff.brain_freeze.react
		Ebonbolt = function()
			return not Talent.GlacialSpike.Enabled()
			    or Player.Buff(Aura.Icicles).Stack() == 5
			   and not Player.Buff(Aura.BrainFreeze).React();
		end,

		Flurry = {
			-- # With GS, the mage only shatters Frostbolt that would put them at 1-3 Icicle stacks, Ebonbolt if it would waste Brain Freeze charge (i.e. when the mage starts casting Ebonbolt with Brain Freeze active) and of course Glacial Spike. Difference between shattering Frostbolt with 1-3 Icicles and 1-4 Icicles is small, but 1-3 tends to be better in more situations (the higher GS damage is, the more it leans towards 1-3).
			-- actions.single+=/flurry,if=talent.glacial_spike.enabled&buff.brain_freeze.react&(prev_gcd.1.frostbolt&buff.icicles.stack<4|prev_gcd.1.glacial_spike|prev_gcd.1.ebonbolt)
			GlacialSpike = function()
				return Talent.GlacialSpike.Enabled()
				   and Player.Buff(Aura.BrainFreeze).React()
				   and (Player.PrevGCD(1, Spell.Frostbolt) and Player.Buff(Aura.Icicles).Stack() < 4 or Player.PrevGCD(1, Talent.GlacialSpike) or Player.PrevGCD(1, Talent.Ebonbolt));
			end,

			-- # Without GS, the mage just tries to shatter as many Frostbolts and Ebonbolts as possible. Forcing shatter on Frostbolt is still a small gain, so is not caring about FoF. Ice Lance is too weak to warrant delaying Brain Freeze Flurry.
			-- actions.single+=/flurry,if=!talent.glacial_spike.enabled&(prev_gcd.1.ebonbolt|buff.brain_freeze.react&prev_gcd.1.frostbolt)
			Use = function()
				return not Talent.GlacialSpike.Enabled()
				   and (Player.PrevGCD(1, Talent.Ebonbolt) or Player.Buff(Aura.BrainFreeze).React() and Player.PrevGCD(1, Spell.Frostbolt));
			end,

			-- actions.single+=/flurry,if=azerite.winters_reach.enabled&!buff.brain_freeze.react&buff.winters_reach.react
			WintersReach = function()
				return Azerite.WintersReach.Enabled()
				   and not Player.Buff(Aura.BrainFreeze).Up()
				   and Player.Buff(Aura.WintersReach).React();
			end,
		},

		-- SimCraft doesn't specify this, but only cast if not moving.
		-- We need to put the condition here because movement check is not working yet.
		-- actions.aoe+=/frostbolt
		Frostbolt = function()
			return not Player.IsMoving();
		end,

		-- # Glacial Spike is used when there's a Brain Freeze proc active (i.e. only when it can be shattered). This is a small to medium gain in most situations. Low mastery leans towards using it when available. When using Splitting Ice and having another target nearby, it's slightly better to use GS when available, as the second target doesn't benefit from shattering the main target.
		-- actions.single+=/glacial_spike,if=buff.brain_freeze.react|prev_gcd.1.ebonbolt|active_enemies>1&talent.splitting_ice.enabled
		GlacialSpike = function(numEnemies)
			return Player.Buff(Aura.BrainFreeze).React()
			    or Player.PrevGCD(1, Talent.Ebonbolt)
			    or numEnemies > 1
			   and Talent.SplittingIce.Enabled();
		end,

		-- # Trying to pool charges of FoF for anything isn't worth it. Use them as they come.
		-- actions.single+=/ice_lance,if=buff.fingers_of_frost.react
		IceLance = function()
			return Player.Buff(Aura.FingersOfFrost).React();
		end,

		-- # In some situations, you can shatter Ice Nova even after already casting Flurry and Ice Lance. Otherwise this action is used when the mage has FoF after casting Flurry, see above.
		-- We don't need to check its off cooldown the addon does that automatically.
		-- actions.single=ice_nova,if=cooldown.ice_nova.ready&debuff.winters_chill.up
		IceNova = function()
			return Target.Debuff(Aura.WintersChill).Up();
		end,

		-- # Ray of Frost is used after all Fingers of Frost charges have been used and there isn't active Frozen Orb that could generate more. This is only a small gain against multiple targets, as Ray of Frost isn't too impactful.
		-- There is no real way to do this in the game, just test if its been less than 10 seconds since cast, its the best we can do.
		-- actions.single+=/ray_of_frost,if=!action.frozen_orb.in_flight&ground_aoe.frozen_orb.remains=0
		RayOfFrost = function()
			return Spell.FrozenOrb.TimeSinceLastUsed() < 10;
		end,
	};

	Objects.FinalizeRequirements(self.Requirements);

	function self.Rotation(action)
		action.EvaluateAction(Talent.IceNova, self.Requirements.IceNova);
		action.EvaluateAction(Spell.Flurry, self.Requirements.Flurry.Use);
		action.EvaluateAction(Spell.Flurry, self.Requirements.Flurry.GlacialSpike);
		-- actions.single+=/frozen_orb
		action.EvaluateAction(Spell.FrozenOrb, true);
		action.EvaluateAction(Spell.Blizzard, self.Requirements.Blizzard.FingersOfFrost);
		action.EvaluateAction(Spell.IceLance, self.Requirements.IceLance);
		-- actions.single+=/comet_storm
		action.EvaluateAction(Talent.CometStorm, true);
		action.EvaluateAction(Talent.Ebonbolt, self.Requirements.Ebonbolt);
		action.EvaluateAction(Talent.RayOfFrost, self.Requirements.RayOfFrost);
		action.EvaluateAction(Spell.Blizzard, self.Requirements.Blizzard.Use);
		action.EvaluateAction(Talent.GlacialSpike, self.Requirements.GlacialSpike);
		-- actions.single+=/ice_nova
		action.EvaluateAction(Talent.IceNova, true);
		action.EvaluateAction(Spell.Flurry, self.Requirements.Flurry.WintersReach);
		action.EvaluateAction(Spell.Frostbolt, self.Requirements.Frostbolt);
		-- actions.single+=/ice_lance
		action.EvaluateAction(Spell.IceLance, true);
	end

	return self;
end

-- Create a variable so we can call the rotations functions.
local rotationSingleTarget = SingleTarget("SingleTarget");

-- Base APL Class
local function APL(rotationName, rotationDescription, specID)
	-- Inherits APL Class so get the base class.
	local self = addonTable.rotationsAPL(rotationName, rotationDescription, specID);

	-- Store the information for the script.
	self.scriptInfo = {
		SpecializationID = self.SpecID,
		ScriptAuthor = "LunaEclipse",
		GuideAuthor = "Malon and SimCraft",
		GuideLink = "https://www.wowhead.com/frost-mage-simple-guide",
		WoWVersion = 80001,
	};

	-- Table to hold requirements to use spells for defensive reasons.
	self.Defensives = {
		IceBarrier = function()
			return not Player.Buff(Aura.IceBarrier).Up()
			   and Player.DamagePredicted(5) >= 15;
		end,

		IceBlock = function()
			return Player.DamagePredicted(5) >= 60;
		end,

		-- Greater invisibility will only be used when solo if you have use stealth while solo checked, because it will drop you out of combat.
		Invisibility = function()
			return (not Player.IsSolo() or Settings.GetCharacterValue("ScriptOptions", "OPT_STEALTH_SOLO") == 1)
			   and Player.DamagePredicted(5) >= 40;
		end,
	};

	-- Table to hold requirements to use spells and items listed in the rotation.
	self.Requirements = {
		-- SimCraft doesn't specify this, but only cast the buff if you don't have it.
		-- actions.precombat+=/arcane_intellect
		ArcaneIntellect = function()
			return not Player.Buff(Aura.ArcaneIntellect).Up();
		end,

		-- # If the mage has FoF after casting instant Flurry, we can delay the Ice Lance and use other high priority action, if available.
		-- actions+=/ice_lance,if=prev_gcd.1.flurry&brain_freeze_active&!buff.fingers_of_frost.react
		IceLance = function()
			return Player.PrevGCD(1, Spell.Flurry)
			   and Player.Buff(Aura.BrainFreeze).Up()
			   and not Player.Buff(Aura.FingersOfFrost).React();
		end,

		-- SimCraft doesn't specify this, but we need to make sure that we don't already have a pet summoned
		-- actions.precombat+=/water_elemental
		SummonWaterElemental = function()
			return not Pet.IsActive();
		end,
	};

	Objects.FinalizeRequirements(self.Defensives, self.Requirements);

	-- Function for setting up action objects such as spells, buffs, debuffs and items, called when the rotation becomes the active rotation.
	function self.Enable(...)
		Racial, Spell, Talent, Aura, Azerite = ...;

		Aura.WintersReach = Objects.newSpell(273347);

		Item = {};

		Consumable = {
			-- Potions
			BattlePotionOfIntellect = Objects.newItem(163222),
		};

		Objects.FinalizeActions(Racial, Spell, Talent, Aura, Azerite, Item, Consumable);
	end

	-- Function for setting up the configuration screen, called when rotation becomes the active rotation.
	function self.SetupConfiguration(config, options)
		config.RacialOptions(options, Racial.AncestralCall, Racial.Berserking, Racial.BloodFury, Racial.Fireblood, Racial.LightsJudgment);
		config.AOEOptions(options, Spell.Blizzard, Talent.CometStorm, Spell.ConeOfCold, Spell.FrozenOrb, Talent.IceNova);
		config.BuffOptions(options, Spell.ArcaneIntellect);
		config.CooldownOptions(options, Talent.Ebonbolt, Talent.GlacialSpike, Spell.IcyVeins, Talent.MirrorImage, Talent.RayOfFrost, Talent.RuneOfPower);
		config.DefensiveOptions(options, Spell.IceBarrier, Spell.IceBlock, Spell.Invisibility);
		config.PetOptions(options, Spell.SummonWaterElemental);
	end

	-- Function for destroying action objects such as spells, buffs, debuffs and items, called when the rotation is no longer the active rotation.
	function self.Disable()
		local coreTables = addon.Core.Tables;

		coreTables.Wipe(Racial, Spell, Talent, Aura, Azerite, Item, Consumable);
	end

	-- Function for checking the rotation that displays on the Defensives icon.
	function self.Defensive(action)
		-- The abilities here should be listed from highest damage required to suggest to lowest,
		-- Specific damage types before all damage types.

		-- Protects against all types of damage
		action.EvaluateDefensiveAction(Spell.IceBlock, self.Defensives.IceBlock);
		action.EvaluateDefensiveAction(Spell.Invisibility, self.Defensives.Invisibility);
		action.EvaluateDefensiveAction(Spell.IceBarrier, self.Defensives.IceBarrier);
	end

	-- Function for displaying opening rotation.
	function self.Opener(action)
	end

	-- Function for displaying any actions before combat starts.
	function self.Precombat(action)
		action.EvaluateAction(Spell.ArcaneIntellect, self.Requirements.ArcaneIntellect);
		action.EvaluateAction(Spell.SummonWaterElemental, self.Requirements.SummonWaterElemental);
		-- actions.precombat+=/mirror_image
		action.EvaluateAction(Talent.MirrorImage, true);
		-- actions.precombat+=/potion
		action.EvaluateAction(Consumable.BattlePotionOfIntellect, true);
		-- actions.precombat+=/frostbolt
		action.EvaluateAction(Spell.Frostbolt, true);
	end

	-- Function for checking the rotation that displays on the Single Target, AOE, Off GCD and CD icons.
	function self.Combat(action)
		action.EvaluateAction(Spell.IceLance, self.Requirements.IceLance);

		-- actions+=/call_action_list,name=cooldowns
		action.CallActionList(rotationCooldowns);
		action.CallActionList(rotationAOE);
		-- actions+=/call_action_list,name=single
		action.CallActionList(rotationSingleTarget);
	end

	return self;
end

local rotationAPL = APL(nameAPL, "LunaEclipse: Frost Mage", addonTable.Enum.SpecID.MAGE_FROST);
