-- Pulls back the Addon-Local Variables and store them locally.
local _, addonTable = ...;
local addon = _G["LunaEclipse_Rotation"];

-- Plugin information variables.
local pluginName = "Weak Auras";

-- Check whether the plugin is supposed to be active and that the required Weak Auras is loaded.
if addon and addon.API.Blizzard.IsAddOnLoaded("WeakAuras") and addon.System.Plugins.IsActive(pluginName) then
	-- Store local copies of Blizzard API and other addon global functions and variables
	local format, ipairs, pairs, print, tinsert, type = format, ipairs, pairs, print, tinsert, type;

	-- Local variables for plugin operation.
	local registeredStyles = {};

	-- Tables to hold private functions.
	local Auras, Backups, Styles = {}, {}, {};

	local function generateDifferences(firstTable, secondTable)
		local Tables = addon.Core.Tables;
		local returnValue = {};

		if Tables.IsTable(firstTable) and Tables.IsTable(secondTable) then
			for key, value in pairs(firstTable) do
				if key ~= "internalVersion" then
					if Tables.IsTable(value) then
						local changes = generateDifferences(value, secondTable[key]);

						if not Tables.IsEmpty(changes) then
							returnValue[key] = changes;
						end
					else
						local secondTableValue = Tables.GetValue(secondTable, key);

						if secondTableValue == nil or (type(value) == "string" and value ~= secondTableValue and value ~= format("%s\n", secondTableValue)) or (type(value) ~= "string" and value ~= secondTableValue) then
							returnValue[key] = secondTable[key];
						end
					end
				end
			end
		end

		if Tables.IsEmpty(returnValue) then
			returnValue = nil;
		end

		return returnValue;
	end

	local function getAuraNames()
		local coreTables = addon.Core.Tables;

		local returnValue = {};
		local addedNames = {};
		local aura, auraID, childAuras;

		for _, styleData in pairs(registeredStyles) do
			aura = coreTables.GetValue(styleData, "Parent", {});
			childAuras = coreTables.GetValue(styleData, "Imports", {});
			auraID = coreTables.GetValue(aura, "id") or coreTables.GetValue(aura.ID, "id");

			if auraID and not coreTables.GetValue(addedNames, auraID) then
				addedNames[auraID] = true;
				tinsert(returnValue, auraID);
			end

			if not coreTables.IsEmpty(childAuras) then
				for _, childAura in ipairs(childAuras) do
					auraID = coreTables.GetValue(childAura, "id") or coreTables.GetValue(childAura.ID, "id");

					if auraID and not coreTables.GetValue(addedNames, auraID) then
						addedNames[auraID] = true;
						tinsert(returnValue, auraID);
					end
				end
			end
		end

		return returnValue;
	end

	local function weakAurasLoaded()
		local coreTables = addon.Core.Tables;
		local auraFound = false;
		local auraNames = getAuraNames();

		if not coreTables.IsEmpty(auraNames) then
			for _, auraID in ipairs(auraNames) do
				local auraData = WeakAuras.GetData(auraID);
				local region = WeakAuras.GetRegion(auraID, "");

				if not coreTables.IsEmpty(auraData) and not coreTables.IsEmpty(region) then
					return true;
				end

				if not auraFound and not coreTables.IsEmpty(auraData) then
					auraFound = true;
				end
			end
		end

		return not auraFound;
	end

	--------------------- Private Module Functions ---------------
	Auras = {
		Add = function(aura)
			local Tables = addon.Core.Tables;

			if not Tables.IsEmpty(aura) then
				local parentName = Tables.GetValue(aura, "parent", "");
				local parentData = parentName ~= "" and WeakAuras.GetData(parentName);

				WeakAuras.Add(aura);
				WeakAuras.SyncParentChildRelationships(true);

				if WeakAuras.OptionsFrame then
					WeakAuras.NewDisplayButton(aura);

					if parentData then
						local button = WeakAuras.GetDisplayButton(parentName);

						if button then
							button.callbacks.UpdateExpandButton();
						end

						WeakAuras.UpdateDisplayButton(parentData);
						WeakAuras.ReloadGroupRegionOptions(parentData);
					end
				end
			end
		end,

		Build = function(aura, styleName)
			local Tables = addon.Core.Tables;
			local auraID = Tables.GetValue(aura, "id") or Tables.GetValue(aura.ID, "id");
			local auraData = WeakAuras.GetData(auraID);
			local auraDefault = Auras.Default(aura);

			if auraID then
				local customAuras = Tables.GetValue(addonTable.weakAuraSettings.CustomAuras, styleName, {});
				local customAura = Tables.Merge(auraDefault, Tables.GetValue(customAuras, auraID));

				if not Tables.IsEmpty(auraData) then
					return Auras.Modify(auraData, aura);
				elseif not Tables.IsEmpty(customAura) then
					return customAura;
				else
					return auraDefault;
				end
			end

			return nil;
		end,

		Default = function(aura)
			local Tables = addon.Core.Tables;
			local returnValue;

			if not Tables.IsEmpty(aura) then
				local templates = addonTable.WeakAuras.Templates;
				local auraType = Auras.Type(aura);

				if auraType == 1 then
					returnValue = templates.templateGroup;
					returnValue = Tables.Merge(returnValue, aura);
				elseif auraType ~= 0 then
					if auraType == 2 then
						returnValue = templates.templateIcon;
					else
						returnValue = templates.templateText;
					end

					returnValue = Tables.Merge(returnValue, aura.ID);
					returnValue = Tables.Merge(returnValue, aura.Editable);
					returnValue = Tables.Merge(returnValue, aura.Uneditable);
				end
			end

			return returnValue;
		end,

		Delete = function(aura)
			local Tables = addon.Core.Tables;

			if not Tables.IsEmpty(aura) then
				if WeakAuras.OptionsFrame then
					local parentName = Tables.GetValue(aura, "parent");

					WeakAuras.DeleteOption(aura);

					if parentName then
						local parentData = WeakAuras.GetData(parentName);
						local parentButton = WeakAuras.GetDisplayButton(parentName);

						if parentData then
							WeakAuras.UpdateGroupOrders(parentData);
						end

						if parentButton then
							parentButton.callbacks.UpdateExpandButton();
						end
					end
				else
					WeakAuras.Delete(aura);
					WeakAuras.SyncParentChildRelationships(true);
				end
			end
		end,

		Modify = function(aura, updateAura)
			local Tables = addon.Core.Tables;
			local returnValue;

			if not Tables.IsEmpty(aura) then
				returnValue = Tables.Duplicate(aura);

				if not Tables.IsEmpty(updateAura) then
					returnValue = Tables.Merge(aura, updateAura.Uneditable);
				end
			end

			return returnValue;
		end,

		Type = function(aura)
			local Tables = addon.Core.Tables;
			local regionType = Tables.GetValue(aura, "regionType") or Tables.GetValue(aura.ID, "regionType", "");

			if regionType == "group" then
				return 1;
			elseif regionType == "icon" then
				return 2;
			elseif regionType == "text" then
				return 3;
			end

			return 0;
		end,
	};

	Backups = {
		Create = function(aura, styleName)
			local Tables = addon.Core.Tables;

			local auraID = Tables.GetValue(aura, "id") or Tables.GetValue(aura.ID, "id");
			local auraData = WeakAuras.GetData(auraID);
			local auraDefault = Auras.Default(aura);

			if not Tables.IsEmpty(auraData) then
				local profile = Tables.GetValue(addonTable.weakAuraSettings.CustomAuras, styleName, {});
				local differences = generateDifferences(auraDefault, auraData);

				-- Aura was modified so be sure to back it up
				profile[auraID] = differences;

				if Tables.IsEmpty(profile) then
					profile = nil;
				end

				-- Make sure we save any aura customization to saved variables
				addonTable.weakAuraSettings.CustomAuras[styleName] = profile;
			end
		end,

		Delete = function(aura, styleName)
			local Tables = addon.Core.Tables;
			local auraID = Tables.GetValue(aura, "id") or Tables.GetValue(aura.ID, "id");

			local profile = Tables.GetValue(addonTable.weakAuraSettings.CustomAuras, styleName, {});

			profile[auraID] = nil;

			if Tables.IsEmpty(profile) then
				profile = nil;
			end

			-- Make sure we save any aura customization to saved variables
			addonTable.weakAuraSettings.CustomAuras[styleName] = profile;
		end,
	};

	Styles = {
		-- Returns false if the delete would fail, this is when there is no region data available.
		AbandonStyle = function(parentName, auraNames)
			local Tables = addon.Core.Tables;

			if parentName and not Tables.IsEmpty(auraNames) then
				local Core = addonTable.WeakAuras.Core;
				local region = WeakAuras.regions[parentName] or {};

				if not Tables.IsEmpty(region.ControlChildren) then
					region:Pause();
				end

				for _, weakAura in ipairs(auraNames) do
					-- We are removing an aura just using a name no longer attached to a style, so just pass the name as a string
					Core.RemoveAura(weakAura);
				end

				-- We are removing an aura just using a name no longer attached to a style, so just pass the name as a string
				Core.RemoveAura(parentName);
			end
		end,

		Apply = function(styleName)
			if styleName then
				local Core = addonTable.WeakAuras.Core;
				local Tables = addon.Core.Tables;

				local addonSettings = addonTable.addonSettings or {};
				local styleData = Tables.GetValue(registeredStyles, styleName);

				if not Tables.IsEmpty(styleData) then
					local parentData = Tables.GetValue(styleData, "Parent");
					local auraImports = Tables.GetValue(styleData, "Imports");

					if not Tables.IsEmpty(parentData) and not Tables.IsEmpty(auraImports) then
						Core.AddAura(parentData, styleName);

						for _, weakAura in ipairs(auraImports) do
							Core.AddAura(weakAura, styleName);
						end

						addonSettings.versionWeakAura = addonTable.WeakAuras.Version;
					end
				end
			end
		end,

		Change = function(newStyle, oldStyle)
			if newStyle ~= nil and oldStyle ~= nil and newStyle ~= oldStyle then
				Styles.RemoveStyle(oldStyle)
				Styles.Apply(newStyle);
			end
		end,

		Register = function(styleName, weakAurasOptions, parentAura, auraImports)
			if styleName then
				local Tables = addon.Core.Tables;
				local styleData = Tables.GetValue(registeredStyles, styleName);

				if Tables.IsEmpty(styleData) then
					registeredStyles[styleName] = {
						Options = weakAurasOptions,
						Parent = parentAura,
						Imports = auraImports,
					};
				end
			end
		end,

		Remove = function(styleName)
			if styleName then
				local Tables = addon.Core.Tables;
				local currentStyleName = Tables.GetValue(addonTable.addonSettings, "styleSelection");

				if styleName ~= currentStyleName then
					Styles.RemoveStyle(styleName)
				end
			end
		end,

		-- Returns false if the delete would fail, this is when there is no region data available.
		RemoveStyle = function(styleName, overrideBackup)
			if styleName then
				local Core = addonTable.WeakAuras.Core;
				local Tables = addon.Core.Tables;

				local styleData = Tables.GetValue(registeredStyles, styleName);

				if not Tables.IsEmpty(styleData) then
					local parentData = Tables.GetValue(styleData, "Parent");
					local auraImports = Tables.GetValue(styleData, "Imports");

					if not Tables.IsEmpty(parentData) and not Tables.IsEmpty(auraImports) then
						local auraID = Tables.GetValue(parentData, "id");
						local region = WeakAuras.regions[auraID] or {};

						if not Tables.IsEmpty(region.ControlChildren) then
							region:Pause();
						end

						for _, weakAura in ipairs(auraImports) do
							Core.RemoveAura(weakAura, styleName, overrideBackup);
						end

						Core.RemoveAura(parentData, styleName, overrideBackup);
					end
				end
			end
		end,

		Verify = function(styleName)
			if styleName then
				local Tables = addon.Core.Tables;
				local styleData = Tables.GetValue(registeredStyles, styleName);

				if styleName == "Hekili" then
					local styleAuras = { "LERH_Defensive", "LERH_CD", "LERH_OffGCD", "LERH_Interrupt", "LERH_PRIORITY_1", "LERH_PRIORITY_2", "LERH_PRIORITY_3", "LERH_PRIORITY_4",
										 "LERH_Bloodlust", "LERH_Range", "LERH_Enemies", "LERH_TimeToDie", "LERH_DamageAll", "LERH_DamageMagic",  "LERH_DamagePhysical" };

					-- Hekili Style is incompatible with new rotation processess, so remove it as our currentStyle.
					Styles.AbandonStyle("LERH_Hekili", styleAuras);

					-- Remove any settings for the Hekili style, and any aura backups.
					addonTable.weakAuraSettings.Styles.Hekili = nil;
					addonTable.weakAuraSettings.CustomAuras.Hekili = nil;

					-- We need to set the style to Ovale which is the default style, and save the change in the configuration.
					styleName = "Ovale";
					addonTable.addonSettings.styleSelection = styleName;

					-- Print a message to the user that the style was removed and reset to Ovale.
					print("LunaEclipse: Rotation Helper - The Hekili WeakAura style is no longer compatible with rotation processing changes and has been removed.");
					print("LunaEclipse: Rotation Helper - Your WeakAura style has now been set to Ovale by default.");

					addonTable.WeakAuras.Core.SetupConfiguration(styleName);
					Styles.Apply(styleName);
				elseif not Tables.IsEmpty(styleData) then
					local parentData = Tables.GetValue(styleData, "Parent");
					local auraID = Tables.GetValue(parentData, "id");
					local parentAura = WeakAuras.GetData(auraID);

					if Tables.IsEmpty(parentAura) or Tables.GetValue(addonTable.addonSettings, "versionWeakAura") < addonTable.WeakAuras.ForcedRefresh then
						-- Either the parent doesn't exist, or its a forced refresh so remove the icons without creating a backup.
						Styles.RemoveStyle(styleName, true);
					end

					Styles.Apply(styleName);
				end
			end
		end,
	};

	------------- Private Internal Namespace Functions -----------
	addonTable.WeakAuras.Core = {
		AddAura = function(aura, styleName)
			if not WeakAuras.IsImporting() and styleName then
				local Tables = addon.Core.Tables;
				local auraID = Tables.GetValue(aura, "id") or Tables.GetValue(aura.ID, "id") or aura;

				if auraID then
					local profile = Tables.GetValue(addonTable.weakAuraSettings.Styles, styleName, {});
					local auraData = WeakAuras.GetData(auraID);
					local newAura = Auras.Build(aura, styleName);
					local auraEnabled = Tables.GetValue(profile, auraID) == 1 or (Tables.GetValue(profile, auraID) == nil and (Auras.Type(aura) == 1 or Auras.Type(aura) == 2));

					if auraEnabled then
						if Tables.IsEmpty(auraData) then
							Auras.Add(newAura);
						else
							if Tables.GetValue(addonTable.addonSettings, "versionWeakAura") < addonTable.WeakAuras.Version then
								Auras.Add(newAura);
							else
								Backups.Create(aura, styleName);
							end
						end
					elseif not Tables.IsEmpty(auraData) then
						Auras.Delete(auraData);
					end
				end
			end
		end,

		-- Returns false if the delete would fail, this is when there is no region data available.
		RemoveAura = function(aura, styleName, overrideBackup)
			local Tables = addon.Core.Tables;

			if not WeakAuras.IsImporting() and aura then
				if styleName then
					local auraID = Tables.GetValue(aura, "id") or Tables.GetValue(aura.ID, "id") or aura;

					if auraID then
						local region = WeakAuras.GetRegion(auraID, "");
						local auraData = WeakAuras.GetData(auraID);

						if not Tables.IsEmpty(region) and not Tables.IsEmpty(auraData) then
							if not overrideBackup then
								Backups.Create(aura, styleName);
							end

							-- Delete the child aura from the group.
							Auras.Delete(auraData);
						end
					end
				else
					local region = WeakAuras.GetRegion(aura, "");
					local auraData = WeakAuras.GetData(aura);

					if not Tables.IsEmpty(region) and not Tables.IsEmpty(auraData) then
						-- Delete the aura.
						Auras.Delete(auraData);
					end
				end
			end
		end,

		SetupConfiguration = function(styleName)
			if styleName then
				local Options = addon.Options.Plugins;
				local Tables = addon.Core.Tables;

				local WeakAurasOptions = addon.Options.pluginOptions.args[Options.CovertPluginName(pluginName)];
				local styleData = Tables.GetValue(registeredStyles, styleName);

				if not Tables.IsEmpty(styleData) then
					local optionsTable = Tables.GetValue(styleData, "Options");

					WeakAurasOptions.args.RotationIcons.args = Tables.Duplicate(optionsTable.RotationIcons.args);
					WeakAurasOptions.args.InfoTexts.args = Tables.Duplicate(optionsTable.InfoTexts.args);

					addon.System.Callbacks.Callback("optionsChanged", "LERH_PluginSettings");
				end
			end
		end,
	};

	---------------------- Public Functions ----------------------
	addon.Features.WeakAuras = {
		ApplyStyle = function(styleName)
			addon.System.Events.CombatEventDelay(Styles.Verify, styleName);
		end,

		ChangeStyle = function(newStyle, oldStyle)
			addon.System.Events.CombatEventDelay(Styles.Change, newStyle, oldStyle);
		end,

		RegisterStyle = function(styleName, weakAurasOptions, parentAura, auraImports)
			Styles.Register(styleName, weakAurasOptions, parentAura, auraImports);
		end,

		RemoveStyle = function(styleName)
			addon.System.Events.CombatEventDelay(Styles.Remove, styleName);
		end,

		VerifyStyle = function(styleName)
			addon.System.Events.CombatEventDelay(Styles.Verify, styleName);
		end,
	};

	local function playerLogin(styleName)
		if not weakAurasLoaded() then
			addon.System.Timers.TimerAdd(playerLogin, 0.25, false, styleName);
		else
			addonTable.WeakAuras.Core.SetupConfiguration(styleName);
			addon.Features.WeakAuras.VerifyStyle(styleName);
		end
	end

	----------------------- Plugin Code Operation ----------------
	local moduleEvents = {
		playerLogin = function()
			local Tables = addon.Core.Tables;

			local addonSettings = addonTable.addonSettings;
			local styleName = Tables.GetValue(addonSettings, "styleSelection", "Ovale");

			playerLogin(styleName);
		end,

		CooldownIcon = function(ActionList)
			WeakAuras.ScanEvents("LERH_CD_UPDATE", ActionList);
		end,

		DefensiveIcon = function(ActionList)
			WeakAuras.ScanEvents("LERH_DEFENSIVE_UPDATE", ActionList);
		end,

		InterruptIcon = function(ActionList)
			WeakAuras.ScanEvents("LERH_INTERRUPT_UPDATE", ActionList);
		end,

		OffGCDIcon = function(ActionList)
			WeakAuras.ScanEvents("LERH_OFFGCD_UPDATE", ActionList);
		end,

		MultiTargetIcon = function(ActionList)
			WeakAuras.ScanEvents("LERH_AOE_UPDATE", ActionList);
		end,

		SingleTargetIcon = function(ActionList)
			WeakAuras.ScanEvents("LERH_ST_UPDATE", ActionList);
		end,

		WeakAurasTexts = function(Enabled)
			WeakAuras.ScanEvents("LERH_TEXTS_UPDATE", Enabled);
		end,
	};

	-- These are only registered if the plugin is active as they control how what is called
	-- from LunaEclipse Rotation Helper
	addon.System.Callbacks.RegisterAddonCallbacks(moduleEvents);
end